from jinja2 import Environment, PackageLoader, select_autoescape

env = Environment(
    loader=PackageLoader('edatk', '_html_report'),
    autoescape=select_autoescape()
)

def _build_template(**kwargs):
    """Render and return html template given keyword args.

    Returns:
        string: rendered html
    """
    template = env.get_template('report_template.html')
    return template.render(**kwargs)


import pytest
import pandas as pd
import seaborn as sns

import warnings
warnings.filterwarnings("ignore", module = "matplotlib\..*" )
warnings.filterwarnings("ignore", module = "seaborn\..*" )
warnings.filterwarnings("ignore", message="FixedFormatter should only be used together with FixedLocator")

import edatk._single_variable._summary_statistics as sst
from edatk._auto_eda import auto_eda


def _get_sns_test_datasets(small_list=True):
    if small_list:
        return [sns.load_dataset('iris'), sns.load_dataset('diamonds'), sns.load_dataset('titanic')]
    else:
        return [sns.load_dataset(ds_name) for ds_name in sns.get_dataset_names()]


def _get_test_df():
    SAMPLE_LIST = [5.1,4.9,4.7,4.6,5.0, None]
    SAMPLE_CAT = ['a', 'b', 'c', 'd', 'e', 'a']
    return pd.DataFrame(list(zip(SAMPLE_LIST,SAMPLE_CAT)), columns=['metric', 'category'])


def test_mean():
    assert round(sst._op_mean(_get_test_df(),'metric'),2) == 4.86


def test_median():
    assert round(sst._op_median(_get_test_df(),'metric'),2) == 4.9


def test_rowcount():
    assert sst._op_rowcount(_get_test_df(),'metric') == 6
    assert type(sst._op_rowcount(_get_test_df(),'metric')) is int


def test_min():
    assert round(sst._op_min(_get_test_df(),'metric'), 2) == 4.6


def test_max():
    assert round(sst._op_max(_get_test_df(),'metric'),2) == 5.1


def test_variance():
    assert round(sst._op_variance(_get_test_df(),'metric'),4) == 0.0344


def test_standard_deviation():
    assert round(sst._op_standard_deviation(_get_test_df(),'metric'),4) == 0.1855


def test_missing():
    assert int(sst._op_missing_rows(_get_test_df(),'metric')) == 1
    assert type(sst._op_missing_rows(_get_test_df(),'metric')) is int


def test_75th_percentile():
    assert round(sst._op_quantile(_get_test_df(),'metric',0.75),4) == 5.0


def test_distinct_count():
    assert int(sst._op_distinct_count(_get_test_df(), 'category')) == 5


def test_data_type():
    assert sst._op_get_column_data_type(_get_test_df(), 'category') == 'string'
    assert sst._op_get_column_data_type(_get_test_df(), 'metric') == 'numeric-condensed'


def test_auto_column_text_eda():
    df = _get_test_df()
    auto_eda(df)


def test_sns_datasets():
    ds_list = _get_sns_test_datasets()
    for i, ds in enumerate(ds_list):
        print(f'Running dataset {i}')
        auto_eda(ds, ignore_errors=False, show_chart=False)



import matplotlib.pyplot as plt
from seaborn.external.docscrape import header

import edatk._core as core
import edatk._single_variable._summary_statistics as sst
import edatk._single_variable._visuals as viz

def _text_box_plot(df, column_name):
    """Return the text box plot given a dataframe and column name string.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be analyzed

    Returns:
        string: box plot as a simple string
    """
    min = sst._op_min(df, column_name)
    tf = sst._op_quantile(df, column_name, 0.25)
    med = sst._op_median(df, column_name)
    sf = sst._op_quantile(df, column_name, 0.75)
    max = sst._op_max(df, column_name)
    try:
        return f'|{min:.2f} --||{tf:.2f} ~ {med:.2f} ~ {sf:.2f}||-- {max:.2f}|'
    except:
        return f'|{min} --||{tf} ~ {med} ~ {sf}||-- {max}|'


def _dist_rank_wrapper(df, column_name, ax):
    _, dist_df = sst._get_theoritical_distributions(df, column_name)
    dist_series = dist_df.set_index('distribution_type')['rmse'].squeeze()
    if dist_series is not None:
        viz._plot_simple_bar(dist_series, f"{column_name} Distribution Fit (RMSE of Density Deltas)", ax)


_auto_eda_column_ops = {
    'numeric': {
        'Column Name': lambda df, column_name: str(column_name),
        'Data Type Grouping': sst._op_get_column_data_type,
        'Data Type': lambda df, column_name: str(df[column_name].dtype),
        'Row Count': sst._op_rowcount,
        'Distinct Count': sst._op_distinct_count,
        'Missing Values': sst._op_missing_rows,
        'Missing Value %': lambda df, column_name: float(sst._op_missing_rows(df, column_name)) / float(sst._op_rowcount(df, column_name)),
        'Mean': sst._op_mean,
        'Median': sst._op_median,
        'Min': sst._op_min,
        'Max': sst._op_max,
        'Standard Deviation': sst._op_standard_deviation,
        'CV %': lambda df, column_name: float(sst._op_standard_deviation(df, column_name)) / float(sst._op_mean(df, column_name)),
        'Skew': sst._op_skew,
        'Kurtosis': sst._op_kurtosis,
        'Text Box Plot': _text_box_plot
    },
    'numeric-condensed': {
        'Column Name': lambda df, column_name: str(column_name),
        'Data Type Grouping': sst._op_get_column_data_type,
        'Data Type': lambda df, column_name: str(df[column_name].dtype),
        'Row Count': sst._op_rowcount,
        'Distinct Count': sst._op_distinct_count,
        'Missing Values': sst._op_missing_rows,
        'Missing Value %': lambda df, column_name: float(sst._op_missing_rows(df, column_name)) / float(sst._op_rowcount(df, column_name)),
        'Mean': sst._op_mean,
        'Median': sst._op_median,
        'Min': sst._op_min,
        'Max': sst._op_max,
        'Standard Deviation': sst._op_standard_deviation,
        'CV %': lambda df, column_name: float(sst._op_standard_deviation(df, column_name)) / float(sst._op_mean(df, column_name)),
        'Skew': sst._op_skew,
        'Kurtosis': sst._op_kurtosis,
        'Text Box Plot': _text_box_plot
    },
    'string': {
        'Column Name': lambda df, column_name: str(column_name),
        'Data Type Grouping': sst._op_get_column_data_type,
        'Data Type': lambda df, column_name: str(df[column_name].dtype),
        'Row Count': sst._op_rowcount,
        'Distinct Count': sst._op_distinct_count,
        'Missing Values': sst._op_missing_rows,
        'Missing Value %': lambda df, column_name: float(sst._op_missing_rows(df, column_name)) / float(sst._op_rowcount(df, column_name))
    },
    'bool': {
        'Column Name': lambda df, column_name: str(column_name),
        'Data Type Grouping': sst._op_get_column_data_type,
        'Data Type': lambda df, column_name: str(df[column_name].dtype),
        'Row Count': sst._op_rowcount,
        'Distinct Count': sst._op_distinct_count,
        'Missing Values': sst._op_missing_rows,
        'Missing Value %': lambda df, column_name: float(sst._op_missing_rows(df, column_name)) / float(sst._op_rowcount(df, column_name))
    }
}

_auto_eda_column_visuals = {
    'numeric': {
        'Box Plot': viz._plot_distributions,
        'Histogram': viz._plot_histogram,
        'Swarm': viz._plot_swarm,
        'ECDF': viz._plot_ecdf,
        'Distributions': viz._plot_distribution_overlay,
        'Best Distribution': lambda df, column_name, ax: viz._plot_distribution_overlay(df, column_name, ax, best_only=True),
        'Distribution Fits': _dist_rank_wrapper
    },
    'numeric-condensed': {
        'Histogram': viz._plot_histogram,
    },
    'string': {
        'Count Plot': viz._plot_categorical_counts,
        'Count Plot %': viz._plot_categorical_percent_counts
    },
    'bool': {
        'Histogram': viz._plot_histogram
    }
}


def _auto_eda_single_column(df, column_name, html_report, show_chart):
    """Print summary statistics and charts given a dataframe and column name string. Ignores NAs besides missing count row.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized
        html_report (object): html report object to hold data and write to file
        show_chart (bool): whether to call plt.show, can be useful to disable in command line interactions
    """
    # Used for separating portions of html doc
    section = 'single_variable'

    # Determine column data type
    data_type = sst._op_get_column_data_type(df, column_name)

    # Try to find operations to do to summarize data type in question
    if data_type in _auto_eda_column_ops:
        column_operations = _auto_eda_column_ops[data_type]
    else:
        error_str = f'{column_name} data type ({data_type}) cannot be parsed.'
        print(error_str)
        if html_report:
            html_report.save_text(error_str, section=section)
        return None

    # Run metric table
    core._bind_to_console_html(section='single_variable', run_type='table', run_dict=column_operations, html_report=html_report, show_chart=show_chart, header_text=column_name, df=df, column_name=column_name)

    # Visual layout
    visual_dict = _auto_eda_column_visuals[data_type]
    core._bind_to_console_html('single_variable', 'charts', visual_dict, html_report, show_chart=show_chart, df=df, column_name=column_name)


def _single_col_ops_error_wrap(df, col, html_report, show_chart):
    section = 'single_variable'
    try:
        _auto_eda_single_column(df, col, html_report, show_chart)
    except:
        error_str = f'{col} was not able to be profiled due to errors'
        print(error_str)
        if html_report:
            html_report.save_text(error_str, section=section)
        plt.close('all')


def _auto_eda_columns(df, column_list=None, html_report=None, ignore_errors=True, show_chart=True):
    """Print summary statistics and charts given a dataframe and list of column name strings. Ignores NAs besides missing count row.

    Args:
        df (pandas dataframe): input dataframe
        column_list (list): list of column names to be summarized
        html_report (HTMLReport class): html report object to write data to
        ignore_errors (bool): whether to ignore errors and run what is possible or raise excpetions
        show_chart (bool): whether to call plt.show, can be useful to disable in command line interactions
    """

    # Check if user pased in list
    if column_list is not None: 
        # Single column
        if isinstance(column_list, str) and column_list in df.columns:
            if ignore_errors:
                _single_col_ops_error_wrap(df, column_list, html_report, show_chart)
            else:
                _auto_eda_single_column(df, column_list, html_report, show_chart)
        else:
            # Multiple defined columns
            for col in column_list:
                if ignore_errors:
                    _single_col_ops_error_wrap(df, col, html_report, show_chart)
                else:
                    _auto_eda_single_column(df, col, html_report, show_chart)
    else:
        # Run all columns
        for col in df.columns:
            if ignore_errors:
                _single_col_ops_error_wrap(df, col, html_report, show_chart)
            else:
                _auto_eda_single_column(df, col, html_report, show_chart)



from itertools import combinations
from pandas.api.types import is_numeric_dtype

import edatk._core as core
import edatk._multi_variable._visuals as viz


def _get_column_combinations(df, column_list=None, target_column=None):
    """Given a dataframe, get a list of column combinations in tuples

    Args:
        df (pd.DataFrame): dataframe to analyze
        column_list (list[str], optional): List of columns to filter combinations down to
        target_column (str, optional): String name of the target column. If None, returns combinations on all columns

    Returns:
        list[tuple[str, str]]: List of column name tuples
    """
    # Filter down columns if needed
    if column_list:
        col_list_init = list(df.columns.values)
        col_list = [col for col in col_list_init if col in column_list]
    else:
        col_list = list(df.columns.values)

    # Filter list to target column in combination if needed
    init_list = list(combinations(col_list, 2))
    if target_column:
        final_list = []
        for combo in init_list:
            col_a, col_b = combo
            if (col_a == target_column) or (col_b == target_column):
                final_list.append(combo)
    else:
        final_list = init_list
    return final_list


def _bind_chart_function(arg_func, **kwargs):
    """For a given set of two columns, wrap those columns as parms into the plot relationship function

    Args:
        func: func to wrap with normal df and ax inputs
        kwargs: passed along to charting function in addition to df and ax

    Returns:
        function: wrapped inner function
    """
    def inner_func(df, ax):
        return arg_func(df=df, ax=ax, **kwargs)
    return inner_func


def _auto_eda_mutli_variable(df, column_list=None, target_column=None, html_report=None, ignore_errors=True, show_chart=True):
    _relationship_ops = {}
    _heatmap_ops = {}
    
    # Get column combination tuples
    column_combinations = _get_column_combinations(df, column_list=column_list)
    target_only_combinations = _get_column_combinations(df, column_list=column_list, target_column=target_column)
    # Loop through tuples
    for col_set in column_combinations:
        # Parse tuple
        col_a, col_b = col_set
        # Enclose function with tuple (df and ax is populated by caller)
        _relationship_ops[f'{col_a}-{col_b}'] = _bind_chart_function(viz._plot_relationship, column_name_one=col_a, column_name_two=col_b, target_column=target_column)
    
    # Run all pair chart functions
    core._bind_to_console_html(section='multi_variable', run_type='charts', run_dict=_relationship_ops, html_report=html_report, show_chart=show_chart, header_text="Column Relationships", df=df)

    # Run heatmap
    # Check for numeric columns
    if column_list is None:
        column_list = df.columns.values
    numeric_col_count = len([col for col in df.columns if col in column_list and is_numeric_dtype(df[col])])
    if numeric_col_count > 0:
        # Standard heatmap
        _heatmap_ops = {
            'Correlation Heatmap': _bind_chart_function(viz._plot_heatmap, column_list=column_list)
        }
        # Target heatmap
        if target_column:
            if is_numeric_dtype(df[target_column]):
                _heatmap_ops['Target Heatmap'] = _bind_chart_function(viz._plot_heatmap, column_list=column_list, target_column=target_column)
        # Visualize all
        for k,v in _heatmap_ops.items():
            core._bind_to_console_html(section='multi_variable', run_type='chart', run_dict={k:v}, html_report=html_report, show_chart=show_chart, header_text=k, df=df)



from edatk._single_variable._auto_eda_single_variable import _auto_eda_columns
from edatk._multi_variable._auto_eda_multi_variable import _auto_eda_mutli_variable
from edatk._single_variable._cardinality_reduction import _add_low_cardinality_target_column
from edatk._core import _check_for_pandas_df
import edatk._html_report._report_builder as html_build


def auto_eda(df, column_list=None, target_column=None, target_low_cardinality_visuals=3, save_path=None, ignore_errors=True, show_chart=True):
    """Run auto eda on a dataframe

    Args:
        df (pd.DataFrame): input dataframe
        column_list (list, optional): List of columns, if none runs for all. Defaults to None.
        target_column (str, optional): Column name string of target. If none runs pair plots for all, otherwise runs only against target. Defaults to None.
        target_low_cardinality_visuals (int, optional): Cardinality of additional target column (if specified) to be added to visualizations where appropriate. Defaults to 3.
        save_path (str, optional): Directory to save html report to. If none then results printed to console instead. Defaults to None.
        ignore_errors (bool, optional): Ignores errors and runs as much as possible. Defaults to True.
        show_chart (bool, optional): Display charts, likely always want this as True unless testing or working with very large datasets. Defaults to True.
    """
     # Initiate html file ops if needed
    if save_path:
        html_report = html_build.HTMLReport(save_path)
    else:
        html_report = None

    # Error checking
    df2 = df.copy()
    _check_for_pandas_df(df2)

    # Grab column list if not passed
    if column_list is None:
        column_list = list(df2.columns)

    # Add new target column for large cardinality
    if target_column is not None:

        # Add an additional target column that is low cardinality for visualization
        _add_low_cardinality_target_column(df=df2, target_column=target_column, desired_cardinality=target_low_cardinality_visuals)
            
        # Check that target is in column list
        if target_column not in column_list:
            column_list.append(target_column)

    # Run single column to console and bind to html if needed
    _auto_eda_columns(df=df2, column_list=column_list, html_report=html_report, ignore_errors=ignore_errors, show_chart=show_chart)

    # Run multi column
    df2.head()
    _auto_eda_mutli_variable(df=df2, column_list=column_list, target_column=target_column, html_report=html_report, ignore_errors=ignore_errors, show_chart=show_chart)

    # Save off final html template
    if html_report:
        html_report.build_final_template()

    # Clean up
    del df2



import math
import numpy as np
import pandas as pd
import seaborn as sns
from seaborn.miscplot import palplot
from seaborn.palettes import color_palette
from edatk._single_variable._summary_statistics import _op_get_column_data_type, _op_distinct_count


def _plot_relationship(df, column_name_one, column_name_two, ax, target_column=None):
    """Plot relationship columns given df and two column names

    Args:
        df (pd.DataFrame): input dataframe
        column_name_one (str): name of column 1 to be compared.
        column_name_two (str): name of column 2 to be compared
        ax (matplotlib ax): chart to plot to.
    """
    # Determine data types of two cols
    dt_one = _op_get_column_data_type(df, column_name_one)
    dt_two = _op_get_column_data_type(df, column_name_two)

    # Set col names as title
    ax.set_title(f'{column_name_one}-{column_name_two}')

    # Hue color column
    if target_column is not None:
        if (target_column != column_name_one) and (target_column != column_name_two):
            hue_color_column = f'{target_column}_lc'
            hue_order = df[hue_color_column].value_counts().index.to_list()
            colors = ['tab:orange', 'tab:purple', 'tab:cyan']
            colors = colors[:len(hue_order)]
            color_palette = {hue:color for hue, color in zip(hue_order, colors)}
        else:
            hue_color_column = None
            hue_order = None
            color_palette = sns.color_palette('tab10')
    else:
        hue_color_column = None
        hue_order = None
        color_palette = sns.color_palette('tab10')

    # --Both numeric = scatter--
    if dt_one == 'numeric' and dt_two == 'numeric':
        
        # X and y fix at zero if possible
        force_axis_to_zero = False # could be later exposed in api
        if force_axis_to_zero: 
            xmax = math.ceil(np.max(df[column_name_one].dropna()) * 1.25)
            xmin = math.floor(np.min(df[column_name_one].dropna()))
            if xmin > 0.0:
                xmin = 0
            ymax = math.ceil(np.max(df[column_name_two].dropna()) * 1.25)
            ymin = math.floor(np.min(df[column_name_two].dropna()))
            if ymin > 0.0:
                ymin = 0
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)

        # Plot scatter
        max_sample = 1000
        if len(df) <= max_sample:
            sns.scatterplot(data=df, x=column_name_one, y=column_name_two, hue=hue_color_column, hue_order=hue_order, palette=color_palette, ax=ax)
        else:
            ct = sns.scatterplot(data=df.sample(n=max_sample, random_state=42), x=column_name_one, y=column_name_two, hue=hue_color_column, hue_order=hue_order, palette=color_palette, ax=ax)
            ct.set_title(f'{ct.get_title()}, n={max_sample} of {len(df)}')
        

    # --Numeric and Condensed = split a kde plot--
    elif (dt_one == 'numeric' and dt_two == 'numeric-condensed') or (dt_one == 'numeric-condensed' and dt_two == 'numeric'):
        if dt_one == 'numeric-condensed':
            string_col = column_name_one
            numeric_col = column_name_two
        else:
            string_col = column_name_two
            numeric_col = column_name_one
        
        sns.kdeplot(data=df, x=numeric_col, hue=string_col, common_norm=False, ax=ax)

    # --Text/Bool as one column, countplot/boxplot/kdeplot depending on data types and unique counts--
    elif (dt_one in ['string', 'bool'] and dt_two in ['numeric', 'numeric-condensed']) or (dt_one in ['numeric', 'numeric-condensed'] and dt_two in ['string', 'bool']):
        if dt_one in ['string', 'bool']:
            string_col = column_name_one
            numeric_col = column_name_two
        else:
            string_col = column_name_two
            numeric_col = column_name_one

        # Replace topn with other
        topn=4
        group_df = df.groupby(string_col)[numeric_col].agg(['sum', 'count']).reset_index()
        top_categories = group_df.sort_values(by='count', ascending=False)[:topn][string_col].astype('str').values.tolist()
        df2 = df.copy()
        df2[string_col] = df2[string_col].astype('str')
        df2[string_col].fillna('Missing', inplace=True)
        top_categories.append('Missing')
        df2.loc[~df2[string_col].isin(top_categories),string_col] = 'Other'

        # Get count of string categories
        distinct_string_count = _op_distinct_count(df2, string_col)

        # Plot distributions with boxplot (string, numeric) or countplot (string, bool)
        if (dt_one == 'bool' or dt_two == 'bool') and (dt_one != 'numeric' and dt_two != 'numeric'):
            sns.countplot(y=df2[numeric_col], hue=df2[string_col], ax=ax)
        elif (dt_one == 'bool' or dt_two == 'bool' or distinct_string_count == 2) and (dt_one == 'numeric' or dt_two == 'numeric'):
            sns.kdeplot(data=df2, x=numeric_col, hue=string_col, common_norm=False, ax=ax)
        else:

            # Box plot categories vs. numeric
            sort_order = df2[string_col].value_counts().reset_index()['index'].values.tolist()
            cpalette = ['red' if 'Missing' in x else 'tab:blue' if 'Other' in x else 'grey' for x in sort_order]
            sns.boxplot(data=df2, x=string_col, y=numeric_col, palette=cpalette, order=sort_order, ax=ax)
            
            # Swarm plot target
            max_sample = 75
            if len(df) > max_sample:
                if hue_color_column is not None:
                    num_groups = _op_distinct_count(df, hue_color_column)
                    group_sample = int(max_sample / num_groups * 1.0)
                    ct = sns.swarmplot(data=df.groupby(hue_color_column).apply(lambda x: x.sample(group_sample, replace=True, random_state=42)), x=string_col, y=numeric_col, hue=hue_color_column, hue_order=hue_order, palette=color_palette, order=sort_order, ax=ax)
                    ct.set_title(f'{ct.get_title()}, swarm n={group_sample * num_groups} (with resampling) of {len(df)}')
                else:
                    ct = sns.swarmplot(data=df2.sample(n=max_sample, random_state=42), x=string_col, y=numeric_col, hue=hue_color_column, hue_order=hue_order, order=sort_order, ax=ax)
                    ct.set_title(f'{ct.get_title()}, swarm n={max_sample} of {len(df)}')
            else:
                sns.swarmplot(data=df2, x=string_col, y=numeric_col, hue=hue_color_column, hue_order=hue_order, order=sort_order, ax=ax)

    # --Both Text/Bool = countplot--
    elif dt_one in ['string', 'bool', 'numeric-condensed'] and dt_two in ['string', 'bool', 'numeric-condensed']:

        # Replace topn with other
        topn=2
        group_df = df.groupby(column_name_one)[column_name_one].agg(['count']).reset_index()
        top_categories = group_df.sort_values(by='count', ascending=False)[:topn][column_name_one].astype('str').values.tolist()
        df2 = df.copy()
        df2[column_name_one] = df2[column_name_one].astype('str')
        df2[column_name_one].fillna('Missing', inplace=True)
        top_categories.append('Missing')
        df2.loc[~df2[column_name_one].isin(top_categories),column_name_one] = 'Other'

        group_df_2 = df.groupby(column_name_two)[column_name_two].agg(['count']).reset_index()
        top_categories_2 = group_df_2.sort_values(by='count', ascending=False)[:topn][column_name_two].astype('str').values.tolist()
        df2[column_name_two] = df2[column_name_two].astype('str')
        df2[column_name_two].fillna('Missing', inplace=True)
        top_categories_2.append('Missing')
        df2.loc[~df2[column_name_two].isin(top_categories_2),column_name_two] = 'Other'
        
        # Plot distributions with barplot (target) or countplot (default)
        df2['combinations'] = df2[column_name_one] + '\n' + df2[column_name_two]
        sort_order = df2['combinations'].value_counts().reset_index()['index'].values.tolist()
        cpalette = ['red' if 'Missing' in x else 'tab:blue' if 'Other' in x else 'grey' for x in sort_order]
        _ = [label.set_fontsize(10) for label in ax.get_yticklabels()]
        if hue_color_column:
            grouped_df = df2.groupby(hue_color_column)['combinations'].value_counts(normalize=True).rename('percentage').reset_index()
            sns.barplot(data=grouped_df, y='combinations', x='percentage', orient='h', hue=hue_color_column, hue_order=hue_order, palette=color_palette, ax=ax)
        else:
            sns.countplot(data=df2, y='combinations', hue=hue_color_column, hue_order=hue_order, order=sort_order, ax=ax)


def _plot_heatmap(df, ax, column_list=None, target_column=None):
    """Plot a heatmap of columns. If target passed, then one col heatmap.

    Args:
        df (pd.DataFrame): input dataframe
        ax (matplotlib ax): ax to plot to
        column_list (str, optional): Columns to be analyzed. Defaults to None.
        target_column (str, optional): Name of target column. Defaults to None.
    """
    if column_list is None:
        column_list = df.columns.values
    
    if target_column:
        ct = sns.heatmap(df.loc[:,column_list].corr()[[target_column]].sort_values(by=target_column, ascending=False), ax=ax, vmin=-1, vmax=1, annot=True, cmap='Spectral')
    else:
        mask = np.triu(np.ones_like(df.loc[:,column_list].corr(), dtype=np.bool))
        ct = sns.heatmap(df.loc[:,column_list].corr(), ax=ax, vmin=-1, vmax=1, annot=True, cmap='Spectral', mask=mask)

    # Fix label rotation
    ct.set_yticklabels(ct.get_yticklabels(), rotation=0)


import numpy as np
import pandas as pd
from pandas.api.types import is_string_dtype, is_numeric_dtype, is_bool_dtype, is_categorical_dtype
import scipy.stats as stats

def _op_mean(df, column_name):
    """Return the numpy mean given a dataframe and column name string. Ignores NAs.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized

    Returns:
        float: mean value
    """
    return np.nanmean(df[column_name])


def _op_median(df, column_name):
    """Return the numpy median given a dataframe and column name string. Ignores NAs.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized

    Returns:
        float: median value
    """
    return np.nanmedian(df[column_name])


def _op_rowcount(df, column_name):
    """Return the numpy row shape given a dataframe and column name string.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized

    Returns:
        int: row count
    """
    return np.array(df[column_name]).shape[0]


def _op_min(df, column_name):
    """Return the numpy min given a dataframe and column name string. Ignores NAs.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized

    Returns:
        float: min value
    """
    return np.nanmin(df[column_name])


def _op_max(df, column_name):
    """Return the numpy max given a dataframe and column name string. Ignores NAs.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized

    Returns:
        float: min value
    """
    return np.nanmax(df[column_name])


def _op_variance(df, column_name):
    """Return the numpy varaiance given a dataframe and column name string. Ignores NAs.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized

    Returns:
        float: variance
    """
    return np.nanvar(df[column_name])


def _op_standard_deviation(df, column_name):
    """Return the numpy standard deviation given a dataframe and column name string. Ignores NAs.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized

    Returns:
        float: standard deviation
    """
    return np.nanstd(df[column_name])


def _op_missing_rows(df, column_name):
    """Return the pandas NA values given a dataframe and column name string.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized

    Returns:
        int: number of rows with missing value
    """
    return int(np.sum(pd.isna(df[column_name])))


def _op_quantile(df, column_name, quantile_value=0.75):
    """Return the quantile (0 to 1 percentile) values given a dataframe and column name string. Ignores NAs.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized

    Returns:
        float: quantile cuttoff point
    """
    if is_bool_dtype(df[column_name]):
        return None
    else:
        return np.nanquantile(df[column_name],quantile_value)


def _op_distinct_count(df, column_name):
    """Return the distinct count given a dataframe and column name string. Ignores NAs.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized

    Returns:
        int: number of unique values
    """
    return df[column_name].nunique()


def _op_skew(df, column_name):
    """Return the scipy skew given a dataframe and column name string. Ignores NAs.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized

    Returns:
        float: skew
    """
    return stats.skew(df[column_name])


def _op_kurtosis(df, column_name):
    """Return the scipy kurtosis given a dataframe and column name string. Ignores NAs.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized

    Returns:
        float: kurtosis
    """
    return stats.kurtosis(df[column_name])


def _op_get_column_data_type(df, column_name):
    """Return the data type given a dataframe and column name string.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be analyzed

    Returns:
        string: data type of the column
    """
    if is_string_dtype(df[column_name]):
        return 'string'
    elif is_bool_dtype(df[column_name]):
        return 'bool'
    elif is_numeric_dtype(df[column_name]):
        if _op_distinct_count(df, column_name) <= 10:
            if _op_distinct_count(df, column_name) == 2:
                if _op_min(df, column_name) == 0 and _op_max(df, column_name) == 1:
                    return 'bool'
            return 'numeric-condensed'
        else:
            return 'numeric'
    elif is_categorical_dtype(df[column_name]):
        return 'string'
    else:
        return str(df[column_name].dtype)


def _get_theoritical_distributions(df, column_name):
    """Compare frequencies of column against theoritical freqencies to determine best fit distribution.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be analyzed

    Returns:
        pandas dataframe: dataframe containing distribtion points
        pandas dataframe: dataframe containing distribution and rmse, sorted descending 
    """
    # Init df concat list to hold distributions and remove NAs from column
    df_concat_list = []
    col_wo_nas = df[column_name].dropna()

    # Calculate the number of bins using numpy defaults
    num_bins = len(np.histogram_bin_edges(np.array(col_wo_nas), bins='auto'))

    # Get histogram from dataframe
    y, x = np.histogram(col_wo_nas, bins=num_bins, density=True)
    midpoints = (x[:-1] + x[1:]) / 2.0

    # Add original distribution to the df concat list
    y_df = pd.DataFrame(y, columns=['distribution']).set_index(pd.Series(midpoints))
    y_df['distribution_type'] = 'original data'
    df_concat_list.append(y_df)

    # Loop through distributions
    dist_list = ['norm', 'expon', 'uniform', 'lognorm', 't']
    for dist_name in dist_list:

        # Fit Dist
        dist = getattr(stats.distributions, dist_name)
        dist_parms = dist.fit(col_wo_nas)

        # Generate pdf from modpoints and distribution parms
        pdf = dist.pdf(midpoints, *dist_parms)

        # Add results to df concat list
        df_dist = pd.DataFrame(pdf, columns=['distribution']).set_index(pd.Series(midpoints))
        df_dist['distribution_type'] = dist_name
        df_concat_list.append(df_dist)


    # Combine into one df
    all_distributions = pd.concat(df_concat_list).reset_index().rename(columns={'index': column_name})

    # Split data by orig and distributions
    orig_data_idx = all_distributions['distribution_type'] == 'original data'
    df_orig_data = all_distributions.loc[orig_data_idx, :]
    df_other_data = all_distributions.loc[~orig_data_idx]

    # Calculate sse
    combined_df = pd.merge(df_other_data, df_orig_data, how='left', on=column_name, suffixes=['_dist','_orig'])
    combined_df['squared_errors'] = (combined_df['distribution_dist'] - combined_df['distribution_orig']) ** 2
    df_se = pd.DataFrame(combined_df.groupby(['distribution_type_dist'])['squared_errors'].mean().reset_index())
    df_se['rmse'] = np.sqrt(df_se['squared_errors']) 
    df_se.drop(['squared_errors'], axis=1, inplace=True)
    df_se = df_se.rename(columns={'distribution_type_dist': 'distribution_type'})
    df_se = df_se.sort_values(by='rmse', ascending=True)

    return all_distributions, df_se


import pathlib
import os
import webbrowser
import glob
from edatk._html_report._template_ops import _build_template

class HTMLReport:
    """Class for capturing html details and rendering + saving file.
    """
    def __init__(self, save_path=None):
        """Create new instance of HTML Report

        Args:
            save_path (string, optional): Path to save off finalized html report and all assets. Defaults to None.
        """
        self._create_report_directory(save_path)
        self._single_variable_charts = []
        self._multi_variable_charts = []


    def _create_report_directory(self, save_path, remove_old_files=True):
        """Create report repository at path

        Args:
            save_path (string): Path to save off finalized html report and all asets
            remove_old_files (bool): Whether to remove old edatk_ files in html directory
        """
        # Calc Save path
        if save_path:
            save_path = os.path.join(save_path, 'html_report')
        else:
            save_path = os.path.join(os.getcwd(), 'html_report')
        
        # Create parent rerport path
        pathlib.Path(save_path).mkdir(parents=True, exist_ok=True)

        # Create underlying assets path
        pathlib.Path(os.path.join(save_path, 'assets')).mkdir(parents=True, exist_ok=True)

        # Save off paths
        self.root_path =  save_path
        self.asset_path = os.path.join(save_path, 'assets')

        # Remove old files if needed
        if remove_old_files:
            del_string = os.path.join(save_path, 'assets', 'edatk_*.png')
            files = glob.glob(del_string)
            for file in files:
                os.remove(file)


    def save_title(self, title, section):
        """Save title to html rendering.

        Args:
            title (string): Text to render as title.
            section (string): Section to bind to.
        """
        title = title.replace("\n",' <br> ')
        if section == 'single_variable':
            self._single_variable_charts.append({'render_type':'title', 'render_value': title})
        elif section == 'multi_variable':
            self._multi_variable_charts.append({'render_type':'title', 'render_value': title})
        

    def save_text(self, text, section):
        """Save text to html rendering.

        Args:
            text (string): Text to render as paragraph.
            section (string): Section to bind to.
        """
        text = text.replace("\n",' <br> ')
        if section == 'single_variable':
            self._single_variable_charts.append({'render_type':'text', 'render_value': text})
        elif section == 'multi_variable':
            self._multi_variable_charts.append({'render_type':'text', 'render_value': text})


    def save_chart_to_image(self, fig, chart_name, section):
        """Save chart to html rendering.

        Args:
            fig (matplotlib fig): Fig to save as png.
            chart_name (string): Chart name to use as file name.
            section (string): Section to bind to.
        """

        # Save off chart
        image_file_name = f'{chart_name}.png'
        fig.savefig(os.path.join(self.asset_path, image_file_name))

        # Add to render pipeline
        if section == 'single_variable':
            self._single_variable_charts.append({'render_type': 'png', 'render_value': image_file_name})
            self._single_variable_charts.append({'render_type': 'lb', 'render_value': 'None'})
        elif section == 'multi_variable':
            self._multi_variable_charts.append({'render_type': 'png', 'render_value': image_file_name})
            self._multi_variable_charts.append({'render_type': 'lb', 'render_value': 'None'})


    def save_table(self, table_list_of_dict, section):
        """Save table to html rendering.

        Args:
            table_list_of_dict (list of dictionary objects): metric, value combination
            section (string): Section to bind to.
        """
        if section == 'single_variable':
            self._single_variable_charts.append({'render_type':'table', 'render_value': table_list_of_dict})
        elif section == 'multi_variable':
            self._multi_variable_charts.append({'render_type':'table', 'render_value': table_list_of_dict})

    
    def build_final_template(self, open_template=True):
        """Build final template and write to file

        Args:
            open_template (bool): Whether final html template file should be opened after building.
        """
        template = _build_template(single_variable_charts=self._single_variable_charts, multi_variable_charts=self._multi_variable_charts)
        write_path = os.path.join(self.root_path, 'report.html')
        with open(write_path, 'w') as f:
            f.write(template)
        print(f'Open web view of this report at {write_path}.')

        if open_template:
            webbrowser.open_new_tab(write_path)


import edatk._single_variable._summary_statistics as sst


def _top_value_or_other(x, top_value_list):
    if x in top_value_list:
        return x
    else:
        return 'Other'


def _range_classifier(x, low, high, column_name):
    if x < low:
        return f'low ({column_name}<{float(round(low,2))})'
    elif x < high:
        return f'medium ({float(round(low,2))}<{column_name}<{float(round(high,2))})'
    else:
        return f'high ({column_name}>{float(round(high,2))})'


def _get_column_reduced_cardinality(df, column_name, cardinality, desired_cardinality):
    """Return a series that is the reduced cardinality version of a supplied df[column_name]

    Args:
        df (pd.DataFrame): input df
        column_name (str): string column name to reduce cardinality on
        cardinality (int): cardinality of the input column
        desired_cardinality (int): desired cardinality

    Returns:
        pd.Series: reduced cardinality pandas series
    """
    if cardinality <= desired_cardinality:
        return df[column_name]
    else:
        target_dtype = sst._op_get_column_data_type(df, column_name)
        s = df[column_name].dropna()
        reduced_cardinality_series = None

        if target_dtype == 'string':
            # Count all values (presort)
            vcounts = s.value_counts()
            top_values = vcounts[:(desired_cardinality-1)]
            reduced_cardinality_series = s.apply(lambda x: _top_value_or_other(x, top_values))

        elif 'numeric' in target_dtype:
            mean_value = sst._op_mean(df, column_name)
            std = sst._op_standard_deviation(df, column_name)
            low = mean_value - std
            high = mean_value + std
            reduced_cardinality_series = s.apply(lambda x: _range_classifier(x, low, high, column_name))

        else:
            reduced_cardinality_series = 'NA'

        return reduced_cardinality_series


def _add_low_cardinality_target_column(df, target_column, desired_cardinality):
    """Add an additional low cardinality derived column in place.

    Args:
        df (pd.DataFrame): input dataframe
        target_column (str): string name of the target column
        desired_cardinality (int): desired cardinality (numeric will always be 3 though).
    """
    
    # Get information about the target column
    inferred_col_name = f'{target_column}_lc'
    target_distinct_count = sst._op_distinct_count(df, target_column)
    
    # Cardinality > desired, reduce cardinality
    if target_distinct_count > desired_cardinality:
        df[inferred_col_name] = _get_column_reduced_cardinality(
            df=df, 
            column_name=target_column, 
            cardinality=target_distinct_count, 
            desired_cardinality=desired_cardinality
        )
    else:
        # If cardinality <= desired, leave as is
        df[inferred_col_name] = df[target_column]


import seaborn as sns
import numpy as np
import pandas as pd
from pandas.api.types import is_bool_dtype
import matplotlib.ticker as mtick
import math

from edatk._core import _rotate_x_axis_labels, _integer_y_axis_format
from edatk._single_variable._summary_statistics import _op_missing_rows as na_rows
from edatk._single_variable._summary_statistics import _get_theoritical_distributions


def _split_top_others(s, topn=10, na_row_count=None):
    """Split a pandas series by topn and others value counts.

    Args:
        s (pandas series): series containing categories (strings).
        topn (int, optional): number to include as individual items, grouping others into one All Other. Defaults to 10.

    Returns:
        pandas series: pandas series with value counts
    """
    # Edge case of under topn values
    if len(s) <= topn:
        return s.value_counts()

    # Count all values (presort) to be split
    vcounts = s.value_counts()

    # Grab slice for top values
    top_values = vcounts[:topn]
    
    # Build series for other values, with combined counts
    other_values_sum = np.sum(vcounts[topn:].values)
    other_value_dict = {
        'Other': other_values_sum
    }
    other_value_series = pd.Series(other_value_dict)

    # Add in NA rows if requested
    missing_value_dict = {
        'Missing': na_row_count
    }
    missing_value_series = pd.Series(missing_value_dict)

    # Return combined values series
    return pd.concat([top_values, other_value_series, missing_value_series])


def _get_percentage_from_counts(vcounts):
    """Given value counts, get the corresponding percentages.

    Args:
        vcounts (pandas series): pandas value counts

    Returns:
        pandas series: value counts as a percentage of total
    """
    return vcounts / np.sum(vcounts)


def _annotate_bars(ax, colors, force_int=False):
    """For a bar chart drawn to ax, annotate labels at top of bar with same colors.

    Args:
        ax (matplotlib ax object): ax object with bars to be annotated
        colors (list): list of color strings
        force_int (bool): force lables to round to nearest int or not
    """
    for p, c in zip(ax.patches, colors):
        if force_int:
            format_str = int(round(p.get_height(),0))
        else:
            format_str = "%.2f" % p.get_height()
        ax.annotate(format_str, (p.get_x() + p.get_width() / 2., p.get_height()),
                    ha='center', va='center', fontsize=11, color=c, xytext=(0, 20),
                    textcoords='offset points')


def _plot_distributions(df, column_name, ax):
    """Return boxplot ax given a dataframe and column name string. Ignores NAs.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized
        ax (matplotlib ax object): ax to plot chart on
    """
    if not is_bool_dtype(df[column_name]):
        # Filter out nas and grab correct column
        filtered_col = df[column_name].dropna()

        # Plot and clean up chart formatting
        ct = sns.boxplot(x=filtered_col, ax=ax)
        ct.set_title(f'{column_name} Box Plot')
        ct.set(xlabel=None)


def _plot_categorical_counts(df, column_name, ax):
    """Plot bars with counts of the various values in the column.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized
        ax (matplotlib ax object): ax to plot chart on
    """

    # Filter out nas and translate topn
    na_row_count = na_rows(df, column_name)
    filtered_col = df[column_name].dropna()
    summarized_col = _split_top_others(filtered_col, topn=5, na_row_count=na_row_count)

    # Change y axis to integer format and pad
    ymax = math.ceil(np.max(summarized_col) * 1.25)
    ymin = math.floor(np.min(summarized_col))
    if ymin > 0.0:
        ymin = 0
    ax.set_ylim(ymin, ymax)
    _integer_y_axis_format(ax)
    
    # Fix x axis labels from overlapping
    _rotate_x_axis_labels(ax)
    
    # Calc color palette
    cpalette = ['tab:blue' if x == 'Other' else 'red' if x == 'Missing' else 'grey' for x in summarized_col.index]

    # Plot chart
    sns.barplot(x=summarized_col.index, y=summarized_col, ax=ax, palette=cpalette).set_title(f'{column_name} Count Plot')

    # Add labels
    _annotate_bars(ax, cpalette, force_int=True)


def _plot_simple_bar(s, title, ax):
    """Plot simple bars from series.

    Args:
        s (pandas series): input dataframe
        title (string): title to add to plot
        ax (matplotlib ax object): ax to plot chart on
    """

    # Filter out nas
    summarized_col = s.dropna()

    # Y axis pad
    ymax = np.max(summarized_col) * 1.25
    ymin = math.floor(np.min(summarized_col))
    if ymin > 0.0:
        ymin = 0
    ax.set_ylim(ymin, ymax)
    
    # Fix x axis labels from overlapping
    _rotate_x_axis_labels(ax)

    # Calc color pallete
    cpalette = ['grey' for x in summarized_col]
    
    # Plot chart
    ct = sns.barplot(x=summarized_col.index, y=summarized_col, ax=ax, palette=cpalette)
    ct.set_title(title)
    ct.set(xlabel=None)
    ct.set(ylabel=None)

    # Add labels
    _annotate_bars(ax, cpalette, force_int=False)


def _plot_categorical_percent_counts(df, column_name, ax):
    """Plot bars with count percents of the various values in the column.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized
        ax (matplotlib ax object): ax to plot chart on
    """

    # Filter out nas and translate to topn percent of total counts
    na_row_count = na_rows(df, column_name)
    filtered_col = df[column_name].dropna()
    summarized_col = _get_percentage_from_counts(_split_top_others(filtered_col, topn=5, na_row_count=na_row_count))
    summarized_col *= 100.0

    # Change y axis to percent format
    ax.yaxis.set_major_formatter(mtick.PercentFormatter())

    # Pad y axis
    ymax = math.ceil(np.max(summarized_col) * 1.25)
    ymin = math.floor(np.min(summarized_col))
    if ymin > 0.0:
        ymin = 0
    ax.set_ylim(ymin, ymax)
    
    # Fix x axis labels from overlapping
    _rotate_x_axis_labels(ax)
    
    # Calc color palette
    cpalette = ['tab:blue' if x == 'Other' else 'red' if x == 'Missing' else 'grey' for x in summarized_col.index]

    # Plot chart
    sns.barplot(x=summarized_col.index, y=summarized_col, ax=ax, palette=cpalette).set_title(f'{column_name} % Count Plot')

    # Add labels
    _annotate_bars(ax, cpalette)


def _plot_histogram(df, column_name, ax):
    """Plot histogram.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized
        ax (matplotlib ax object): ax to plot chart on
    """

    # Plot chart and clean up formatting
    ct = sns.histplot(data=df[column_name].dropna(), kde=True, ax=ax)
    ct.set_title(f'{column_name} Histogram')
    ct.set(xlabel=None)
    ct.set(ylabel=None)


def _plot_distribution_overlay(df, column_name, ax, best_only=False):
    """Plot distribution overlay.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized
        ax (matplotlib ax object): ax to plot chart on
        best_only (bool): whether to plot the best fit only or all distributions
    """

    # Calculate various lines for distribution
    all_distributions, rankings = _get_theoritical_distributions(df, column_name)

    # filter to just one if needed
    if best_only:
        best_text = rankings['distribution_type'].values[0]
        best_idx = all_distributions['distribution_type'] == best_text
        default_idx = all_distributions['distribution_type'] == 'original data'
        all_distributions = all_distributions[best_idx | default_idx]

    # Plot chart and clean up formatting
    ct = sns.lineplot(data=all_distributions, x=column_name, y='distribution', hue='distribution_type', ax=ax)
    if best_only:
        ct.set_title(f'{column_name} Closest Distribution')
    else:
        ct.set_title(f'{column_name} Distribution Overlays')
    ct.set(xlabel=None)
    ct.set(ylabel=None)


def _plot_ecdf(df, column_name, ax):
    """Plot ecdf.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized
        ax (matplotlib ax object): ax to plot chart on
    """
    ct = sns.ecdfplot(data=df[column_name].dropna(), ax=ax)
    ct.set_title(f'{column_name} ECDF')
    ct.set(xlabel=None)
    ct.set(ylabel=None)


def _plot_swarm(df, column_name, ax):
    """Plot swarmplot.

    Args:
        df (pandas dataframe): input dataframe
        column_name (string): column name to be summarized
        ax (matplotlib ax object): ax to plot chart on
    """
    # Sampling
    s = df[column_name].dropna()
    max_sample = 1000
    if len(s) <= max_sample:
        max_sample = len(df)
        sample_string = ''
        t=s
    else:
        t = s.sample(n=max_sample, random_state=42)
        sample_string = f' {max_sample} max sample'

    # Plot chart and clean up formatting
    ct = sns.swarmplot(y=t, ax=ax)
    ct.set_title(f'{column_name} Swarmplot{sample_string}')
    ct.set(xlabel=None)
    ct.set(ylabel=None)


from setuptools import setup


with open("README.md", "r") as readme_file:
    readme = readme_file.read()


requirements = [
    'seaborn>=0.11',
    'numpy>=1.20',
    'matplotlib>=3.4.2',
    'pandas>=1.2.4',
    'jinja2>=3.0.1',
    'scipy>=1.6.3'
]

setup(
    name="edatk",
    version="0.0.6",
    author="Barrett Studdard",
    author_email="barrettstuddard@gmail.com",
    description="edatk: python exploratory data analysis toolkit",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/edatk/edatk/",
    packages=['edatk', 'edatk._html_report', 'edatk._single_variable', 'edatk._multi_variable'],
    package_data={'edatk': ['_html_report/*.html']},
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)


import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import pandas as pd
import math
from datetime import datetime


def _check_for_pandas_df(df):
    """Check that input is a pandas dataframe

    Args:
        df (pandas dataframe): object to check instance against
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"


def _get_fig_size_dynamic(num_plots, columns):
    """Calculate x and y figure size based on plots and columns.

    Args:
        num_plots (int): number of visual plots in total.
        columns (int): number of columns that will be displayed.

    Returns:
        tuple: calculated x and y size in float format.
    """
    current_fig_size = plt.rcParamsDefault['figure.figsize']
    x, y = current_fig_size
    return 11.0, y * num_plots / float(columns)


def _get_rows_calc(num_plots, columns):
    """Calculate number of rows for subplots given total plots and columns.

    Args:
        num_plots (int): number of visual plots in total.
        columns (int): number of columns that will be displayed.

    Returns:
        int: number of rows to use.
    """
    rows = math.ceil(num_plots / float(columns))
    return rows


def get_fig_ax(total_num_plots, columns=2):
    """Get fig, axs, and row/column dict given total plots and number of columns.

    Args:
        total_num_plots (int): number of visual plots in total.
        columns (int, optional): number of columns that will be displayed. Defaults to 2.

    Returns:
        tuple: fig, axs, row_col_dict
    """

    # Rows and columns from plot count
    if total_num_plots > 1:
        rows = _get_rows_calc(total_num_plots, columns)
    else:
        rows = 1
        columns = 1

    # Figsize override
    fig_size_d = _get_fig_size_dynamic(rows * columns, columns)

    # Get fig and axs given rows and cols
    sns.set_theme()
    sns.set_style('darkgrid')
    fig, axs = plt.subplots(rows, columns, squeeze=False, figsize=fig_size_d)
    sns.despine(left=True, bottom=True) # must be done after fig to avoid printing dims
    plt.tight_layout(pad=5.0)

    # For number of charts and dims, generate row col tuples
    row_col_dict = {}
    for i in range(rows * columns):
        row = int(i // columns)
        column = int(i % columns)
        row_col_dict[i] = (row, column)

    # Turn off last axis if odd number and > 1
    if (total_num_plots % columns != 0) and (total_num_plots > 1):
        axs[-1,-1].axis('off')
    
    return fig, axs, row_col_dict


def _rotate_x_axis_labels(ax):
    """Rotate the x axis labels slightly to prevent overlapping.

    Args:
        ax (matplotlib ax object): ax object to have x labels rotated
    """
    ax.set_xticklabels(ax.get_xticklabels(),rotation=30)


def _integer_y_axis_format(ax):
    """Force y axis into integer format.

    Args:
        ax (matplotlib ax object): ax object to force integers on y axis
    """
    ax.yaxis.set_major_locator(ticker.MaxNLocator(integer=True))


def _bind_to_console_html(section, run_type, run_dict, html_report, show_chart=True, header_text=None, **kwargs):
    """Bind result of a run dict to consule or html

    Args:
        section (string): section grouping, used for html partitioning
        run_type (string): 'table' or 'charts'
        run_dict (dict): dictionary of operation names and functions
        html_report (HTMLReport class): html report instance or None if should just print to console
        show_chart (bool): Whether to show chart or not when running in console mode
        header_text (string): If not none, will print or title with a header text
        **kwargs: any arguments that should be passed into each row in run_dict
    """
    
    # Check for only valid run types
    assert run_type in ['table', 'chart', 'charts'], "Invalid run type, must be table or charts"

    # Operation header
    if header_text:
        if html_report:
            html_report.save_title(header_text, section=section)
        else:
            print('\n')
            header_text = f'========== {header_text} =========='
            print(header_text)

    
    if run_type == 'table':
        
        # Initiate console and html description string
        if html_report:
            table_list_of_dict = []
        else:
            result = ''
        
        # Loop through operations
        for k, op in run_dict.items():
            
            # Execute op
            op_result = op(**kwargs)
            
            # Dynamic format bind
            if isinstance(op_result, str):
                if html_report:
                    table_list_of_dict.append({'metric':k, 'value':op_result})
                else:
                    result += f'{k:20}: {op_result}'
            elif k[-1:] == '%':
                op_result *= 100.0
                if html_report:
                    table_list_of_dict.append({'metric':k, 'value':f'{round(op_result,2)}%'})
                else:
                    result += f'{k:20}: {op_result:.2f}%'
            elif isinstance(op_result, int):
                if html_report:
                    table_list_of_dict.append({'metric':k, 'value':op_result})
                else:
                    result += f'{k:20}: {op_result}'
            elif isinstance(op_result, float):
                if html_report:
                    table_list_of_dict.append({'metric':k, 'value':round(op_result,2)})
                else:
                    result += f'{k:20}: {op_result:.2f}'
            else:
                if html_report:
                    table_list_of_dict.append({'metric':k, 'value':op_result})
                else:
                    result += f'{k:20}: {op_result}'

            # New line for next result
            if not html_report:
                result += '\n'
        
        # save to file if needed or print combined string back to console
        if html_report:
            html_report.save_table(table_list_of_dict, section=section)
        else:
            print(result)


    elif run_type in['chart', 'charts']:
       
        # Visual layout
        fig, axs, row_col_dict = get_fig_ax(len(run_dict), 2)

        # Build visuals
        for i, (k, visual) in enumerate(run_dict.items()):
            # Find chart placement
            row, col = row_col_dict[i]
            ax = axs[row, col]

            # Plot chart
            visual(**kwargs, ax=ax)

    elif run_type == 'chartx':

        # Single chart bind to fig
        for i, (k, visual) in enumerate(run_dict.items()):
            fig = visual(**kwargs)


    if run_type in ['chart', 'charts']:

        # Save figure if needed
        if html_report and fig:
            html_report.save_chart_to_image(fig, f'edatk_{run_type}_{section}_{datetime.utcnow().strftime("%m_%d_%Y_%H_%M_%S_%f")}', section=section)
        else:
            if show_chart and fig:
                plt.show()
        
        # Cleanup    
        plt.close('all')



import warnings
warnings.filterwarnings("ignore", module = "matplotlib\..*" )
warnings.filterwarnings("ignore", module = "seaborn\..*" )
warnings.filterwarnings("ignore", message="FixedFormatter should only be used together with FixedLocator")
warnings.filterwarnings("ignore", message="Converting input from bool to <class 'numpy.uint8'> for compatibility.")

from ._core import get_fig_ax
from ._auto_eda import auto_eda

__all__ = [
    "auto_eda",
    "get_fig_ax"
]

__version__ = '0.0.6'
